import pytest

from pepmeasure.maxquant import _convert_peptide_txt
from pepmeasure.constants import DATA_PATH


@pytest.mark.parametrize(
    "dataset",
    [
        "./data/peptides.txt",
        DATA_PATH / "peptides.txt",
        pytest.param(lambda: open(DATA_PATH / "peptides.txt", "r"), id="text_file"),
        pytest.param(lambda: open(DATA_PATH / "peptides.txt", "rb"), id="binary_file"),
    ],
)
def test_convert_peptide_txt(dataset):
    if callable(dataset):
        with dataset() as f:
            df = _convert_peptide_txt(f)
    else:
        df = _convert_peptide_txt(dataset)

    assert not df.empty
    assert [
        "Sample",
        "Protein ID",
        "Sequence",
        "Intensity",
        "PEP",
        "Charges",
    ] == df.columns.tolist()



import pytest

import pandas as pd
from plotly import exceptions

from pepmeasure.utils import (
    sanitize_seq,
    get_column_name,
    get_distinct_seq,
    normalize_color,
    extract_related_kwargs,
    convert_exponential_to_suffix,
    get_match_for_seq,
)
from tests.constants import PEPTIDES


def test_sanitize_seq():
    assert "PEPTIDE" == sanitize_seq("pEPtiDe :)")


def test_get_column_name():
    assert "Intensity" == get_column_name(PEPTIDES, "intensity")
    normalized = PEPTIDES.rename(columns={"Intensity": "Normalized intensity"})
    assert "Normalized intensity" == get_column_name(normalized, "intensity")
    with pytest.raises(ValueError) as e:
        get_column_name(PEPTIDES, "test")
    assert "could not be found" in str(e.value)


def test_get_distinct_seq():
    peptides = pd.DataFrame(
        {
            "Sequence": [
                "SRVLNLGPITRK",
                "NDPFANKDDPFYYDWKNLQ",
                "SRVLNLGPITRK",
                "EEGEFEEEAEEEVA",
                "GPPGPPGPPGHPGPQGPPG",
                "GPPGPPGPPGHPGPQGPPG",
            ]
        }
    )
    expected = pd.DataFrame(
        {
            "Sequence": [
                "SRVLNLGPITRK",
                "NDPFANKDDPFYYDWKNLQ",
                "EEGEFEEEAEEEVA",
                "GPPGPPGPPGHPGPQGPPG",
            ]
        }
    ).reset_index(drop=True)
    assert expected.equals(get_distinct_seq(peptides).reset_index(drop=True))


def test_normalize_color():
    assert "rgb(13, 8, 135)" == normalize_color(0.0, 0.0, 100.0, "Plasma")
    assert "rgb(182, 48, 139)" == normalize_color(42.0, 0.0, 100.0, "Plasma")
    with pytest.raises(exceptions.PlotlyError) as e:
        normalize_color(42.0, 0.0, 100.0, "Something")
    assert "not a built-in scale" in str(e.value)


def test_extract_related_kwargs():
    mapping = {
        "param_a": "param_1",
        "param_b": "param_2",
        "param_c": "param_3",
    }
    params = {
        "param_a": "foo",
        "param_c": "foo",
        "param_d": "foo",
    }
    expected = {
        "param_1": "foo",
        "param_3": "foo",
    }
    assert expected == extract_related_kwargs(mapping, params)


def test_convert_exponential_to_suffix():
    assert "10" == convert_exponential_to_suffix(1)
    assert "10k" == convert_exponential_to_suffix(4)
    assert "10M" == convert_exponential_to_suffix(7)
    assert "10B" == convert_exponential_to_suffix(10)


def test_get_match_for_seq():
    data = pd.DataFrame(
        {
            "Sample": ["AD01_C1_INSOLUBLE_01", "CTR01_C1_INSOLUBLE_01"],
            "Protein ID": ["A0A075B6S2", "A0A075B6S2"],
            "Sequence": ["FSGVPDR", "PEPTIDE"],
            "Intensity": [936840.0, "NaN"],
            "PEP": [0.0068633, 0.0056387],
            "GRAVY": [2.0, 1.0],
        }
    )
    seq = "PEPTIDE"
    expected_match = {
        "Sequence": "PEPTIDE",
        "Protein ID": "A0A075B6S2",
        "GRAVY": 1.0,
    }
    assert (1, expected_match) == get_match_for_seq(data, seq)
    assert (0, {}) == get_match_for_seq(data, "PEP")



from dataclasses import dataclass
from functools import partial
import os
from pathlib import Path
import pickle
import string
import sys
from typing import Callable

from modlamp.descriptors import GlobalDescriptor
import numpy as np
import pandas as pd
from pyteomics import mass

from pepmeasure.constants import (
    AA_FORMULA,
    AA_LETTERS,
    AA_THREE_LETTER_CODE,
    AA_ONE_LETTER_CODE,
    AA_WEIGHTS,
    HYDROPATHY_INDICES,
    WATER,
    CHEMICAL_CLASS,
    CHARGE_CLASS,
    PROTON_MASS,
)
from pepmeasure.utils import (
    sanitize_seq,
    get_distinct_seq,
    extract_related_kwargs,
)


def _seq_length(seq: str) -> int:
    """
    Computes the length in a given sequence.
    Note: The input sequence must be pre-sanitized to compute only valid amino acids.
        seq: Given sequence
    """
    invalid = set(seq) - AA_LETTERS
    if invalid:
        raise ValueError(f"Invalid amino acid symbol: {', '.join(sorted(invalid))}")
    return len(seq)


def _aa_frequency(seq: str) -> dict[str, int]:
    """
    Computes the frequency of each amino acid in a given sequence.
    Note: The input sequence must be pre-sanitized to compute only valid amino acids.
        seq: Given sequence
    """
    try:
        freq = {val: 0 for val in AA_LETTERS}
        for aa in seq:
            freq[aa] += 1
        return freq
    except KeyError as e:
        raise ValueError(f"Invalid amino acid symbol: '{e.args[0]}'") from None


def _molecular_weight(seq: str) -> float:
    """
    Computes the average molecular weight of a given sequence in Da.
    Note: The input sequence must be pre-sanitized to compute only valid amino acids.
        seq: Given sequence
    """
    num = _seq_length(seq)
    weight = sum(AA_WEIGHTS[aa] for aa in seq) - (num - 1) * WATER
    return round(weight, 2)


def _mass(seq: str, option: str = "monoisotopic") -> float:
    """
    Computes the mass of a given peptide sequence.
        seq: Given sequence
        option: Specification of which mass type to compute, can be "monoisotopic" or "average".
    """
    if option == "monoisotopic":
        return round(mass.calculate_mass(sequence=seq, monoisotopic=True), 2)
    if option == "average":
        return round(mass.calculate_mass(sequence=seq, monoisotopic=False), 2)
    else:
        raise ValueError(f"Unknown option: {option}")


def _theoretical_mz(
    seq: str,
    charges: int | float | list[int] | list[float] | str = [1],
) -> str:
    """
    Computes the mass-to-charge ratio (m/z) for a sequence and its observed charge states.
        seq: Given sequence
        charges: A charge state or a list of charge states.
    """
    # TODO Move this conversion to MaxQuant import function
    try:
        if charges is np.nan:
            charges = [1]  # Default
        elif type(charges) is str:
            charges = [int(z) for z in charges.split(";") if z.strip()]
        elif type(charges) is int or type(charges) is float:
            charges = [charges]
        else:
            charges = [int(z) for z in charges]
    except Exception as e:
        raise ValueError(f"Parameter 'Charge' could not be processed: {e}")

    mz_vals = [
        mass.calculate_mass(sequence=seq, monoisotopic=True, charge=z) for z in charges
    ]
    return ", ".join([f"{mz:.2f}" for mz in mz_vals])


def _three_letter_code(seq: str) -> str:
    """
    Converts a sequence of amino acids into its representation in three letter code.
    Note: The input sequence must be pre-sanitized to compute only valid amino acids.
        seq: Given sequence
    """
    try:
        return "".join(AA_THREE_LETTER_CODE[aa] for aa in seq)
    except KeyError as e:
        raise ValueError(f"Invalid amino acid symbol: '{e.args[0]}'") from None


def _one_letter_code(codes: str) -> str:
    """
    Converts concatenated three-letter amino acid codes into their one-letter code representation.
        codes: Sequence in three letter code
    """
    separators = set(string.whitespace + string.punctuation)
    if any(ch in separators for ch in codes):
        raise ValueError(
            f"Invalid input: Separators between codes are not allowed."
        ) from None
    seq = []
    for i in range(0, len(codes), 3):
        code = codes[i : i + 3]
        try:
            seq.append(AA_ONE_LETTER_CODE[code])
        except KeyError as e:
            raise ValueError(f"Invalid three letter code: '{e.args[0]}'.") from None
    return "".join(seq)


def _gravy(seq: str) -> float:
    """
    Computes the GRAVY (grand average of hydropathy) score of a given sequence.
    Note: The input sequence must be pre-sanitized to compute only valid amino acids.
        seq: Given sequence
    """
    num = _seq_length(seq)
    hydropathy_sum = sum(HYDROPATHY_INDICES[aa] for aa in seq)
    return round(hydropathy_sum / num, 3)


def _molecular_formula(seq: str) -> str:
    """
    Computes the molecular formula of a given sequence.
    Note: The input sequence must be pre-sanitized to compute only valid amino acids.
        seq: Given sequence
    """
    total_atoms = {}
    for aa in seq:
        for atom, count in AA_FORMULA[aa].items():
            total_atoms[atom] = total_atoms.get(atom, 0) + count

    num_bindings = _seq_length(seq) - 1
    total_atoms["H"] -= 2 * num_bindings
    total_atoms["O"] -= num_bindings

    sorted_atoms = ["C", "H", "N", "O", "S"]
    formula_elems = [
        f"{atom}{total_atoms[atom]}" if total_atoms[atom] > 1 else atom
        for atom in sorted_atoms
        if atom in total_atoms
    ]
    return "".join(formula_elems)


def _isoelectric_point(seq: str, option: str = "bjellqvist") -> float:
    """
    Computes the theoretical pI of a given sequence. One can choose between IPC 2.0 (Kozlowski, 2021) to predict
    the pI with the pretrained model IPC2.peptide.svr19 or biopython package based on (Bjellqvist, 1993)
        seq: Given sequence
        option: Specification of which approach to use, can be "bjellqvist" or "kozlowski"
    """
    clean_seq = sanitize_seq(seq)

    if option == "kozlowski":
        EXTERNAL_PATH = Path(__file__).resolve().parent / "external"
        ipc_path = EXTERNAL_PATH / "ipc-2.0.1"
        model_path = ipc_path / "models" / "IPC2_peptide_75_SVR_19.pickle"
        if os.path.exists(ipc_path):
            sys.path.append(str(ipc_path / "scripts"))
        else:
            raise RuntimeError("IPC 2.0 installation could not be found.")

        # Ignoring warning for import from local module
        from ipc2_lib.svr_functions import get_pI_features  # type: ignore

        X, _ = get_pI_features([[clean_seq, ""]])
        X = np.array(X)

        with open(model_path, "rb") as f:
            model = pickle.load(f)

        return float(round(model.predict(X)[0], 2))

    elif option == "bjellqvist":
        desc = GlobalDescriptor(seq)
        desc.isoelectric_point(amide=False)
        return float(round(desc.descriptor[0][0], 2))

    else:
        raise ValueError(f"Unknown option: {option}")


def _aromaticity(seq: str) -> float:
    """
    Computes the aromaticity of a given sequence by calculating the relative
    frequency of amino acids F, Y, and W (Lobry and Gautier, 1994).
        seq: Given sequence
    """
    freq = _aa_frequency(seq)
    seq_len = _seq_length(seq)
    num_aromatic = sum(freq[aa] for aa in ["F", "Y", "W"])
    return round(num_aromatic / seq_len, 3)


def _aa_classification(seq: str, classify_by: str = "chemical") -> dict:
    """
    Computes the absolute frequency of each class (Pommié et al., 2004).
        seq: Given sequence
        classify_by: Specification of how the amino acids should be classified, can be "chemical" or "charge"
    """
    freq = _aa_frequency(seq)
    if classify_by == "chemical":
        return {
            _class: sum(freq[aa] for aa in aminos)
            for _class, aminos in CHEMICAL_CLASS.items()
        }
    elif classify_by == "charge":
        return {
            _class: sum(freq[aa] for aa in aminos)
            for _class, aminos in CHARGE_CLASS.items()
        }
    else:
        raise ValueError(f"Unknown option: {classify_by}")


def _charge_at_ph(seq: str, ph: float = 7.0) -> float:
    """
    Computes the charge of a given sequence at a given pH level.
        seq: Given sequence
        ph: Given ph level
    """
    desc = GlobalDescriptor(seq)
    desc.calculate_charge(ph=ph, amide=False)
    return float(round(desc.descriptor[0][0], 2))


def _charge_density(seq: str, ph: float = 7.0) -> float:
    """
    Computes the charge density (charge / molecular weight) of a given sequence at a given pH level.
        seq: Given sequence
        ph: Given ph level
    """
    return round(_charge_at_ph(seq, ph) / _molecular_weight(seq), 5)


def _boman_index(seq: str) -> float:
    """
    Computes the boman index of a given sequence (Boman, 2003).
        seq: Given sequence
    """
    desc = GlobalDescriptor(seq)
    desc.boman_index()
    return float(round(desc.descriptor[0][0], 2))


def _aliphatic_index(seq: str) -> float:
    """
    Computes the aliphatic index of a given sequence (Ikai, 1980).
        seq: Given sequence
    """
    freq = _aa_frequency(seq)
    length = _seq_length(seq)
    nA = freq["A"]
    nV = freq["V"]
    nI = freq["I"]
    nL = freq["L"]
    return round((nA + 2.9 * nV + 3.9 * (nI + nL)) * 100.0 / length, 2)


def _extinction_coefficient(seq: str, oxidized: bool = False) -> int:
    """
    Computes the extinction coefficient of a given sequence. Formula is based on (Gill, von Hippel, 1989) and improved by (Pace et al., 1995).
        seq: Given sequence
        oxidized: True, if all pairs of Cysteine form cystines (disulfide bridges). False, if all Cysteine residues are reduced.
    """
    freq = _aa_frequency(seq)
    extinction = freq["W"] * 5500 + freq["Y"] * 1490
    if oxidized:
        extinction += (freq["C"] // 2) * 125
    return extinction


def _instability_index(seq: str) -> float:
    """
    Computes the instability index based on (Guruprasad et al., 1990) of a given sequence.
        seq: Given sequence
    """
    desc = GlobalDescriptor(seq)
    desc.instability_index()
    return float(round(desc.descriptor[0][0], 2))


@dataclass
class Feature:
    label: str
    numeric: bool
    method: Callable
    param_map: dict = None


FEATURES = {
    "molecular_weight": Feature("Molecular weight", True, _molecular_weight),
    "three_letter_code": Feature("Three letter code", False, _three_letter_code),
    "molecular_formula": Feature("Molecular formula", False, _molecular_formula),
    "seq_length": Feature("Sequence length", True, _seq_length),
    "aromaticity": Feature("Aromaticity", True, _aromaticity),
    "aliphatic_index": Feature("Aliphatic index", True, _aliphatic_index),
    "charge_at_ph": Feature(
        "Charge", True, _charge_at_ph, {"charge_at_ph_level": "ph"}
    ),
    "charge_density": Feature(
        "Charge density",
        True,
        _charge_density,
        {"charge_density_level": "ph"},
    ),
    "isoelectric_point": Feature(
        "Isoelectric point",
        True,
        _isoelectric_point,
        {"isoelectric_point_option": "option"},
    ),
    "gravy": Feature("GRAVY", True, _gravy),
    "extinction_coefficient": Feature(
        "Extinction coefficient",
        True,
        _extinction_coefficient,
        {"extinction_coefficient_oxidized": "oxidized"},
    ),
    "boman_index": Feature("Boman index", True, _boman_index),
    "instability_index": Feature("Instability index", True, _instability_index),
    "mass": Feature("Mass", True, _mass, {"mass_option": "option"}),
    "theoretical_mz": Feature(
        "Theoretical m/z", False, _theoretical_mz, {"theoretical_mz_charges": "charges"}
    ),
}


def _compute_features(
    params: dict,
    df: pd.DataFrame = None,
    seq: str = None,
) -> pd.DataFrame:
    """
    Computes all selected features on a pandas DataFrame. See API class 'Calculator' for more information.
    """
    select_all = params.get("select_all")
    # On single sequence or dataset
    if seq is not None:
        df = pd.DataFrame({"Sequence": [seq]})
        sequences = pd.DataFrame({"Sequence": [seq]})
    else:
        sequences = get_distinct_seq(df)

    # Feature mappings as (label, function call with optional params)
    mappings = {}
    for key, feature in FEATURES.items():
        kwargs = (
            extract_related_kwargs(feature.param_map, params)
            if feature.param_map
            else {}
        )
        func = feature.method if not kwargs else partial(feature.method, **kwargs)
        mappings[key] = (feature.label, func)
    # Filter selected features (feature = True)
    chosen_features = {
        col: func
        for feature, (col, func) in mappings.items()
        if params.get(feature) or select_all
    }

    # Compute features
    for feature, func in chosen_features.items():
        # TODO Move this logic
        # (Theoretical m/z is the only feature that processes MaxQuant output)
        if feature == "Theoretical m/z":
            if "Charges" in df.columns:
                sequences["Theoretical m/z"] = df.apply(
                    lambda row: _theoretical_mz(
                        seq=row["Sequence"], charges=row["Charges"]
                    ),
                    axis=1,
                )
            else:
                sequences[feature] = sequences["Sequence"].apply(func)
        else:
            sequences[feature] = sequences["Sequence"].apply(func)

    merged = pd.merge(
        df,
        sequences,
        on="Sequence",
        how="left",
    )
    return merged



from dataclasses import dataclass
from typing import Callable
import pandas as pd
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from plotly.colors import sample_colorscale
from plotly.subplots import make_subplots
from scipy.stats import mannwhitneyu
import warnings

from pepmeasure.constants import (
    AA_WEIGHTS,
    COLORS,
    COLORS_BY_NAME,
    HYDROPATHY_INDICES,
    CHEMICAL_CLASS,
    CHARGE_CLASS,
    CHEMICAL_CLASS_PER_AA,
    CHARGE_CLASS_PER_AA,
)
from pepmeasure.features import (
    _aa_frequency,
    _aa_classification,
    _charge_at_ph,
    _seq_length,
)
from pepmeasure.utils import (
    get_column_name,
    normalize_color,
    extract_related_kwargs,
    convert_exponential_to_suffix,
)


# Sequence-based
def _aa_distribution(
    seq: str,
    order_by: str = "frequency",
    show_all: bool = False,
    colors: list = COLORS,
) -> go.Figure:
    """
    Computes a bar plot showing the frequency distribution for a given sequence.
        seq: Given sequence
        order_by: Specification of how the amino acids should be sorted, can be any of "frequency", "alphabetical", "classes chemical", "classes charge", "hydropathy" or "weight".
        show_all: Specification if all amino acids should be listed, even when not found in the sequence
        colors: List of color codes (in hexadecimal format) to use in the plot
    """
    freq = _aa_frequency(seq)
    num = _seq_length(seq)
    df = pd.DataFrame({"Amino acid": freq.keys(), "Frequency": freq.values()})
    if not show_all:
        df = df[df["Frequency"] > 0]
    df["Proportion"] = (df["Frequency"] / num * 100).round(3)
    if order_by in ["frequency", "alphabetical", "hydropathy", "weight"]:
        fig = px.bar(
            df,
            x="Amino acid",
            y="Frequency",
            color_discrete_sequence=colors,
            custom_data=["Proportion"],
        )
        if order_by == "frequency":
            fig.update_xaxes(categoryorder="total ascending")
        elif order_by == "alphabetical":
            fig.update_xaxes(categoryorder="category ascending")
        elif order_by == "hydropathy":
            sorted_aa = sorted(df["Amino acid"], key=lambda aa: HYDROPATHY_INDICES[aa])
            fig.update_xaxes(categoryorder="array", categoryarray=sorted_aa)
        elif order_by == "weight":
            sorted_aa = sorted(df["Amino acid"], key=lambda aa: AA_WEIGHTS[aa])
            fig.update_xaxes(categoryorder="array", categoryarray=sorted_aa)
        fig.update_traces(
            hovertemplate="Amino acid=%{x}<br>Frequency=%{y} (%{customdata} %)<extra></extra>",
        )

    elif order_by in ["classes chemical", "classes charge"]:
        if order_by == "classes chemical":
            df["Class"] = df["Amino acid"].map(CHEMICAL_CLASS_PER_AA)
            classes = list(CHEMICAL_CLASS.keys())
        elif order_by == "classes charge":
            df["Class"] = df["Amino acid"].map(CHARGE_CLASS_PER_AA)
            classes = list(CHARGE_CLASS.keys())

        # Filter occuring classes
        classes = [cls for cls in classes if (df["Class"] == cls).any()]
        # Compute column widths by number of aa per class
        col_widths = [df[df["Class"] == cls].shape[0] for cls in classes]
        rel_column_widths = [w / sum(col_widths) for w in col_widths]

        # Create plot
        fig = make_subplots(
            cols=len(classes),
            rows=1,
            subplot_titles=classes,
            column_widths=rel_column_widths,
            horizontal_spacing=0.3 / len(classes),
            shared_yaxes=True,
        )
        # Create bar per class
        CLASS_TO_COLOR = dict(zip(classes, colors))
        for i, cls in enumerate(classes):
            class_df = df[df["Class"] == cls].sort_values("Amino acid")
            fig.add_trace(
                go.Bar(
                    x=class_df["Amino acid"],
                    y=class_df["Frequency"],
                    marker_color=CLASS_TO_COLOR[cls],
                    hovertemplate="Amino acid=%{x}<br>Frequency=%{y} (%{customdata} %)<extra></extra>",
                    showlegend=False,
                    customdata=class_df["Proportion"],
                ),
                row=1,
                col=i + 1,
            )
            fig.update_xaxes(title_text=None, row=1, col=i + 1, tickfont=dict(size=10))
        fig.update_layout(
            barmode="group",
            yaxis_title="Frequency",
            annotations=[dict(font=dict(size=12)) for _ in fig.layout.annotations],
        )

    else:
        raise ValueError(f"Unknown option for sorting amino acids: {order_by}.")
    fig.update_layout(title=f"Amino acid frequency of sequence {seq}")
    fig.update_yaxes(tickmode="linear", tick0=0, dtick=1)
    return fig


def _hydropathy_profile(seq: str, colors: list = COLORS) -> go.Figure:
    """
    Computes a hydropathy profile plot for a given sequence.
    Note: The input sequence must be pre-sanitized to compute only valid amino acids.
        seq: Given sequence
        colors: List of color codes (in hexadecimal format) to use in the plot
    """
    df = pd.DataFrame({"Amino acid": list(seq)})
    df["Hydropathy index"] = df["Amino acid"].map(HYDROPATHY_INDICES)
    df.index = df.index + 1
    baseline_df = pd.DataFrame(
        {"Amino acid": ["None"], "Hydropathy index": [0.0]}, index=[0]
    )
    df = pd.concat([baseline_df, df])
    df.index.name = "Residue number"

    fig = px.line(
        df,
        y="Hydropathy index",
        title=f"Hydropathy plot of sequence {seq}",
        hover_data={"Amino acid": True, "Hydropathy index": True},
    )
    fig.update_traces(line=dict(color=colors[0], width=3))
    fig.add_hline(
        y=0,
        line_dash="dash",
    )
    return fig


def _classification(
    seq: str, classify_by: str = "chemical", colors: list = COLORS
) -> go.Figure:
    """
    Computes a bar plot showing the frequency of each amino acid class based on (Pommié et al., 2004).
        seq: Given sequence
        classify_by: Specification of how the amino acids should be classified, can be "chemical" or "charge".
        colors: List of color codes (in hexadecimal format) to use in the plot
    """
    classification = _aa_classification(seq, classify_by)
    num = _seq_length(seq)
    df = pd.DataFrame(
        {"Class": classification.keys(), "Frequency": classification.values()}
    )
    df["Proportion"] = (df["Frequency"] / num * 100).round(3)
    fig = px.bar(
        df,
        x="Class",
        y="Frequency",
        title=f"Classification ({classify_by}) of {seq}",
        color="Class",
        color_discrete_sequence=colors,
        custom_data=["Proportion"],
    )
    fig.update_yaxes(tickmode="linear", tick0=0, dtick=1)
    fig.update_traces(
        showlegend=False,
        hovertemplate="Class=%{x}<br>Frequency=%{y} (%{customdata} %)<extra></extra>",
    )
    return fig


# Dataset-wide
def _titration_curve(seq: str, colors: list = COLORS) -> go.Figure:
    """
    Computes a graph showing the net charge of a given sequence per pH level.
        seq: Given sequence
        colors: List of color codes (in hexadecimal format) to use in the plot
    """
    ph_vals = np.arange(0.0, 14.0 + 0.1, 0.1)
    df = pd.DataFrame({"pH": ph_vals})
    df["Charge"] = df["pH"].apply(lambda ph: _charge_at_ph(seq, ph))
    fig = px.line(
        df,
        x="pH",
        y="Charge",
        title="Titration curve (charge vs. pH)",
    )
    fig.update_traces(line=dict(color=colors[0], width=3))
    fig.add_hline(y=0, line_dash="dash")

    min_charge = int(np.floor(df["Charge"].min()))
    max_charge = int(np.ceil(df["Charge"].max()))
    int_charges = np.arange(min_charge, max_charge + 1)

    points = []
    for i in int_charges:
        matched_charges = df[np.isclose(df["Charge"], i, rtol=0.01)]
        if not matched_charges.empty:
            median_ph = matched_charges["pH"].median()
            points.append((median_ph, i))
    if points:
        ph, charge = zip(*points)
        fig.add_trace(
            go.Scatter(
                x=ph,
                y=charge,
                mode="markers",
                marker=dict(size=8, color=colors[1]),
                hovertemplate="pH=%{x:.1f}<br>Charge=%{y:.0f}<extra></extra>",
                showlegend=False,
            )
        )
    return fig


def _compare_features(
    df: pd.DataFrame,
    feature_a: str = "Sequence length",
    feature_b: str = "Molecular weight",
    group_by: str = None,
    intensity_threshold: float = None,
    colors: list = COLORS,
) -> go.Figure:
    """
    Creates a scatter plot to compare two features across a metadata aspect.
        df: Dataframe that contains the features
        group_by: Metadata aspect (e.g. Group, Batch, ...) that peptides get grouped by
        feature_a: Feature shown on x-axis
        feature_b: Feature shown on y-axis
        intensity_threshold: Peptides with intensities below this threshold are not included
        colors: List of color codes (in hexadecimal format) to use in the plot
    """
    if feature_a not in df.columns:
        raise ValueError(
            f"Feature {feature_a} could not be found in dataset. Please make sure to compute it first."
        )
    if feature_b not in df.columns:
        raise ValueError(
            f"Feature {feature_b} could not be found in dataset. Please make sure to compute it first."
        )
    peptides = df.copy()
    intensity_col = get_column_name(peptides, "intensity")
    if intensity_threshold is not None:
        peptides = peptides[peptides[intensity_col] > intensity_threshold]

    fig = px.scatter(
        peptides,
        x=feature_a,
        y=feature_b,
        color=group_by,
        color_discrete_sequence=colors,
        symbol=group_by,
        symbol_sequence=["square", "circle", "arrow-up", "star"],
        title=f"Comparison of peptide features across each {group_by}",
        hover_name="Sequence",
    )
    if feature_b == "GRAVY":
        fig.add_hline(y=0)
    fig.update_traces(marker=dict(size=10))
    return fig


def _compare_feature(
    df: pd.DataFrame,
    feature: str = "Sequence length",
    group_by: str = None,
    intensity_threshold: float = None,
    colors: list = COLORS,
) -> go.Figure:
    """
    Creates box plots for each group to compare a feature between metadata aspect.
        df: Dataframe that contains the features
        group_by: Metadata aspect (e.g. Group, Batch, ...) that peptides get grouped by
        feature: Feature to be compared
        intensity_threshold: Peptides with intensities below this threshold are not included
        colors: List of color codes (in hexadecimal format) to use in the plot
    """
    if feature not in df.columns:
        raise ValueError(
            f"Feature {feature} could not be found in dataset. Please make sure to compute it first."
        )
    peptides = df.copy()
    intensity_col = get_column_name(peptides, "intensity")
    if intensity_threshold is not None:
        peptides = peptides[peptides[intensity_col] > intensity_threshold]

    fig = px.box(
        peptides,
        x=group_by,
        y=feature,
        color=group_by,
        color_discrete_sequence=colors,
        title=f"Distribution of {feature} across each {group_by}",
        hover_name="Sequence",
    )
    return fig


def _raincloud(
    df: pd.DataFrame,
    feature: str = "Sequence length",
    group_by: str = "Group",
    log_scaled: bool = True,
    colors: list = None,
) -> go.Figure:
    """
    Creates a raincloud plot (containing half violin, box and scatter) for displaying
    the intensity distribution as well as a chosen feature value.
        df: Dataframe that contains the features
        group_by: Metadata aspect (e.g. Group, Batch, ...) that peptides get grouped by
        feature: Feature to be shown in scatter plot
        log_scaled: If True, the x-axis uses a logarithmic (log10) transformation of the intensity data.
    """
    intensity_col = get_column_name(df, "intensity")
    if group_by == "Group" and group_by not in df.columns:
        df["Group"] = "None"

    groups = df[group_by].unique()

    # Sizes & spacings
    violin_width = 0.5
    violin_margin_bottom = -0.8
    violin_margin_top = 0.7
    box_width = 0.075
    violin_box_spacing = -0.043
    scatter_min = -0.3
    scatter_max = -0.1

    # Min & max used for unified color scaling
    min_feature_val = df[feature].min()
    max_feature_val = df[feature].max()
    colorscale = "Plasma"

    fig = make_subplots(
        rows=len(groups),
        cols=1,
        shared_xaxes=True,
    )

    for i, group in enumerate(groups):
        peptides = df[(df[group_by] == group) & (df[intensity_col].notna())].copy()
        if log_scaled:
            intensities = np.log10(peptides[intensity_col])
        else:
            intensities = peptides[intensity_col]

        norm_vals = (peptides[feature] - min_feature_val) / max(
            max_feature_val - min_feature_val, 1.0
        )

        peptides["Color"] = sample_colorscale(colorscale, norm_vals)

        violin_y = np.zeros(len(intensities))
        box_y = np.full(len(intensities), violin_box_spacing)
        scatter_y = np.random.uniform(
            low=scatter_min, high=scatter_max, size=len(intensities)
        )

        violin = go.Violin(
            x=intensities,
            y=violin_y,
            orientation="h",
            side="positive",
            width=violin_width,
            box_visible=False,
            points=False,
            showlegend=False,
            fillcolor=COLORS_BY_NAME["lightgray"],
            line=dict(color=COLORS_BY_NAME["lightgray"]),
            hoverinfo="skip",
        )
        scatter = go.Scattergl(
            x=intensities,
            y=scatter_y,
            mode="markers",
            marker=dict(size=5, color=peptides["Color"]),
            showlegend=False,
            text=peptides[feature],
            customdata=np.stack(
                [peptides[intensity_col], peptides["Sequence"]], axis=-1
            ),
            hovertemplate=(
                "Intensity=%{customdata[0]}<br>"
                f"{feature}=%{{text}}<br>"
                "Sequence=%{customdata[1]}<br>"
                f"{group_by}={group}<extra></extra>"
            ),
        )
        box = go.Box(
            x=intensities,
            y=box_y,
            orientation="h",
            whiskerwidth=0.5,
            width=box_width,
            boxpoints=False,
            showlegend=False,
            fillcolor=COLORS_BY_NAME["transparent"],
            line=dict(color=COLORS_BY_NAME["darkgray"]),
            hoverinfo="skip",
        )
        fig.add_trace(violin, row=i + 1, col=1)
        fig.add_trace(scatter, row=i + 1, col=1)
        fig.add_trace(box, row=i + 1, col=1)

        # Add title & set margins by y-range
        fig.update_yaxes(
            row=i + 1,
            col=1,
            title_text=group,
            range=[
                violin_width * violin_margin_bottom,
                violin_width * violin_margin_top,
            ],
            showticklabels=False,
            zeroline=False,
        )

    # Add invisible trace to show shared colorscale
    fig.add_trace(
        go.Scatter(
            x=[None],
            y=[None],
            mode="markers",
            marker=dict(
                colorscale=colorscale,
                cmin=min_feature_val,
                cmax=max_feature_val,
                showscale=True,
                size=0,
                color=[min_feature_val, max_feature_val],
                colorbar=dict(
                    title=feature,
                    outlinewidth=0,
                ),
            ),
            hoverinfo="none",
            showlegend=False,
        )
    )
    # Adjust x-axis ticks to log10
    if log_scaled:
        valid_logscaled = np.log10(df[intensity_col].dropna())
        min_val = int(np.floor(valid_logscaled.min()))
        max_val = int(np.ceil(valid_logscaled.max()))
        tickvals = list(range(min_val, max_val + 1))
        ticktext = [convert_exponential_to_suffix(t) for t in tickvals]
        fig.update_xaxes(
            row=len(groups),
            col=1,
            tickvals=tickvals,
            ticktext=ticktext,
        )

    # Show y-axis title only on last subplot
    fig.update_xaxes(
        row=len(groups),
        col=1,
        title_text="Intensity (log10)" if log_scaled else "Intensity",
    )
    fig.update_layout(title=f"Raincloud: Intensity and {feature} distribution")
    return fig


def _mann_whitney_u_test(
    df: pd.DataFrame,
    feature: str = "GRAVY",
    group_by: str = "Group",
    group_a: str = "",
    group_b: str = "",
    alternative: str = "two-sided",
    intensity_threshold: float = None,
    colors: list = COLORS,
) -> go.Figure:
    """
    Performs a Mann-Whitney U test on a feature between two groups and creates a box plot with a significance bracket and p-value.
        df: pandas DataFrame that contains the features
        feature: Feature to be compared
        group_by: Metadata aspect (e.g. Group, Batch, ...) that peptides get grouped by
        group_a: First comparison group
        group_b: Second comparison group
        alternative: Chosen test alternative (two-sided, greater, less)
        intensity_threshold: Peptides with intensities below this threshold are not included
        colors: List of color codes (in hexadecimal format) to use in the plot
    """
    peptides = df.copy()
    intensity_col = get_column_name(peptides, "intensity")
    if intensity_threshold is not None:
        peptides = peptides[peptides[intensity_col] > intensity_threshold]

    # Prepare data
    if not group_a or not group_b:
        all_groups = peptides[group_by].unique()
        if len(all_groups) < 2:
            raise ValueError(
                f"Not enough options for {group_by} in metadata, but 2 are required."
            )
        group_a, group_b = all_groups[:2]
        warnings.warn(
            f"{group_a} and {group_b} were selected for performing Mann-Whitney U test because at least one input was empty."
        )
        sub = peptides[[group_by, feature, "Sequence"]].copy()
    else:
        sub = peptides.loc[
            peptides[group_by].isin([group_a, group_b]), [group_by, feature, "Sequence"]
        ].copy()
    sub = sub.dropna(subset=[feature])
    sub = sub.drop_duplicates([group_by, "Sequence"], keep="first")
    x = sub.loc[sub[group_by] == group_a, feature].to_numpy()
    y_pos = sub.loc[sub[group_by] == group_b, feature].to_numpy()
    num_x, num_y = len(x), len(y_pos)
    if num_x < 2 or num_y < 2:
        # raise ValueError(
        #     f"Not enough values: {group_a}={num_x}, {group_b}={num_y} (at least 2 per group required)."
        # )
        text = f"Not enough values: {group_a}={num_x}, {group_b}={num_y} (at least 2 per group required after filtering of intensities). "

        fig = go.Figure()

        # Square
        fig.add_shape(
            type="rect",
            x0=0,
            y0=0,
            x1=1,
            y1=1,
            line=dict(color="black", width=2),
            fillcolor="white",
        )

        # Centered wrapped text
        fig.add_annotation(
            x=0.5,
            y=0.5,
            text=text.replace(" ", "<br>", 8),  # simple manual wrap example
            showarrow=False,
            xanchor="center",
            yanchor="middle",
            align="center",
            font=dict(size=14),
        )

        fig.update_layout(
            title=dict(
                text=f"Mann-Whitney U test of <br> {feature}: {group_a} vs {group_b}",
            ),
            width=400,
            height=400,
            margin=dict(l=10, r=10, t=10, b=10),
            xaxis=dict(range=[0, 1], showgrid=False, zeroline=False, visible=False),
            yaxis=dict(
                range=[0, 1],
                showgrid=False,
                zeroline=False,
                visible=False,
                scaleanchor="x",  # ensures square aspect ratio
                scaleratio=1,
            ),
        )
    else:
        # Execute test
        mw = mannwhitneyu(x, y_pos, alternative=alternative, method="auto")
        p = float(mw.pvalue)

        # Boxplot
        fig = px.box(
            sub,
            x=group_by,
            y=feature,
            color=group_by,
            color_discrete_sequence=colors,
            title=f"Mann-Whitney U test of {feature}: {group_a} vs {group_b}",
            hover_name="Sequence",
        )

        # Significance bracket
        BOX_BRACKET_GAP = 0.05
        BRACKET_HEIGHT = 0.03

        feature_min = float(sub[feature].min())
        feature_max = float(sub[feature].max())
        span = max(feature_max - feature_min, 1.0)
        y_pos = feature_max + BOX_BRACKET_GAP * span
        y_height = BRACKET_HEIGHT * span
        x_group_a, x_group_b = group_a, group_b

        fig.add_shape(  # Left part
            type="line",
            xref="x",
            yref="y",
            x0=x_group_a,
            x1=x_group_a,
            y0=y_pos,
            y1=y_pos + y_height,
        )
        fig.add_shape(  # Middle part
            type="line",
            xref="x",
            yref="y",
            x0=x_group_a,
            x1=x_group_b,
            y0=y_pos + (y_height * 0.88),
            y1=y_pos + (y_height * 0.88),
        )
        fig.add_shape(  # Right part
            type="line",
            xref="x",
            yref="y",
            x0=x_group_b,
            x1=x_group_b,
            y0=y_pos + y_height,
            y1=y_pos,
        )
        fig.add_annotation(
            x=0.5,
            xref="paper",
            y=y_pos + y_height,
            yref="y",
            text=f"p = {round(p, 3)}",
            showarrow=False,
            align="center",
            yanchor="bottom",
        )

        fig.update_yaxes(
            range=[
                feature_min - BOX_BRACKET_GAP * span,
                y_pos + y_height + 2 * BOX_BRACKET_GAP * span,
            ]
        )
    return fig


@dataclass
class Plot:
    seq_based: bool
    method: Callable
    param_map: dict = None


PLOTS = {
    # Sequence-based
    "aa_distribution": Plot(
        True,
        _aa_distribution,
        {
            "aa_distribution_order_by": "order_by",
            "aa_distribution_show_all": "show_all",
        },
    ),
    "classification": Plot(
        True,
        _classification,
        {"classification_classify_by": "classify_by"},
    ),
    "hydropathy_profile": Plot(True, _hydropathy_profile),
    "titration_curve": Plot(True, _titration_curve),
    # Dataset-wide
    "compare_features": Plot(
        False,
        _compare_features,
        {
            "compare_features_a": "feature_a",
            "compare_features_b": "feature_b",
            "compare_features_group_by": "group_by",
            "compare_features_intensity_threshold": "intensity_threshold",
        },
    ),
    "compare_feature": Plot(
        False,
        _compare_feature,
        {
            "compare_feature_group_by": "group_by",
            "compare_feature_a": "feature",
            "compare_feature_intensity_threshold": "intensity_threshold",
        },
    ),
    "raincloud": Plot(
        False,
        _raincloud,
        {
            "raincloud_feature": "feature",
            "raincloud_group_by": "group_by",
        },
    ),
    "mann_whitney": Plot(
        False,
        _mann_whitney_u_test,
        {
            "mann_whitney_feature": "feature",
            "mann_whitney_group_by": "group_by",
            "mann_whitney_group_a": "group_a",
            "mann_whitney_group_b": "group_b",
            "mann_whitney_alternative": "alternative",
            "mann_whitney_intensity_threshold": "intensity_threshold",
        },
    ),
}


def _generate_plots(
    seq: str,
    df: pd.DataFrame,
    params: dict,
    colors: list = COLORS,
) -> list:
    """
    Computes all selected plots on a given pandas DataFrame. Returns a tuple of lists, containing the peptide-specific plots and the plots describing the whole dataset.
    """
    seq_plots = []
    data_plots = []
    select_all = params.get("select_all")
    for key, plot in PLOTS.items():
        if params.get(key) or select_all:
            kwargs = (
                extract_related_kwargs(plot.param_map, params) if plot.param_map else {}
            )
            kwargs["colors"] = colors
            if plot.seq_based and seq is not None:
                kwargs["seq"] = seq
                seq_plots.append(plot.method(**kwargs))
            if not plot.seq_based and df is not None:
                kwargs["df"] = df
                data_plots.append(plot.method(**kwargs))
    return seq_plots, data_plots



# ─────────────────────────────────────────────────────
#                     PLAYGROUND
#   This file is just an experimental sandbox and is
#   not part of the implementation.
#
#   To run this code, execute python -m pepmeasure.
# ──────────────────────────────────────────────────────

from pepmeasure import Calculator

Calculator.hello_pepmeasure()



import pandas as pd
from plotly.colors import sample_colorscale

from pepmeasure.constants import AA_LETTERS


def sanitize_seq(seq: str) -> str:
    """
    Converts all letters to upper case and removes any character that
    does not represent an amino acid according to IUPAC-IUB standard.
    """
    seq = seq.upper()
    return "".join(res for res in seq if res in AA_LETTERS)


def get_column_name(df: pd.DataFrame, keyword: str) -> str:
    """
    Finds the first column of a DataFrame that contains a given keyword.
    """
    col_name = next((col for col in df.columns if keyword.lower() in col.lower()), None)
    if col_name is None:
        raise ValueError(
            f"Keyword {keyword.lower()} could not be found in containing columns: {df.columns}."
        )
    return col_name


def get_distinct_seq(df: pd.DataFrame) -> pd.DataFrame:
    """
    Returns a pandas DataFrame containing only unique sequences.
    """
    return df[["Sequence"]].drop_duplicates(keep="first")


def normalize_color(
    val: float, min: float, max: float, colorscale: str = "Plasma"
) -> float:
    """
    Normalizes a feature value to [0,1] and maps it to a color from a given Plotly colorscale.
        val: Feature value to normalize
        min: Minimum feature value
        max: Maximum feature value
        colorscale: Name of a Plotly colorscale (see https://plotly.com/python/builtin-colorscales/ for more information)
    """
    norm = (val - min) / (max - min)
    return sample_colorscale(colorscale, norm)[0]


def extract_related_kwargs(mapping: dict, params: dict) -> dict:
    """
    Extracts entries whose external key appears in a given mapping from a dictionary. The found entries are returned with their internal key and values with None are ignored.
        mapping: Dictionary which maps external keys to internal keys
        params: Dictionary containing external keys and values
    """
    kwargs = {}
    for external_key, internal_key in mapping.items():
        val = params.get(external_key)
        if val is not None:
            kwargs[internal_key] = val
    return kwargs


def convert_exponential_to_suffix(exp: int) -> str:
    """
    Converts an exponent (e.g., 6 for 10e6) to a human-readable suffix (e.g., 1M)
    """
    value = 10**exp
    if value >= 10**12:
        return f"{value//10**12}T"
    elif value >= 10**9:
        return f"{value//10**9}B"
    elif value >= 10**6:
        return f"{value//10**6}M"
    elif value >= 10**3:
        return f"{value//10**3}k"
    else:
        return str(int(value))


def get_match_for_seq(data: pd.DataFrame, seq: str) -> dict:
    """
    Matches the given sequence to a row of computed sequences and removes all columns that do not correspond to a feature.
    Returns the number of matches and the found features as dict.
    """
    from pepmeasure.features import FEATURES

    ALLOWED = {f.label for f in FEATURES.values()} | {"Sequence", "Protein ID"}
    matched = data[data["Sequence"] == seq]
    num_matches = len(matched)
    matched = matched.loc[:, matched.columns.isin(ALLOWED)]
    if not matched.empty:
        return (num_matches, matched.iloc[0].to_dict())
    else:
        return (0, {})



import pytest
import plotly.graph_objects as go

from pepmeasure import Calculator
from pepmeasure.features import FEATURES
from pepmeasure.plots import PLOTS
from tests.constants import PEPTIDES, METADATA
from pepmeasure.constants import DATA_PATH


def test_simple_init():
    calc = Calculator(
        dataset=PEPTIDES,
        metadata=METADATA,
        seq="SVIDQSRVLNLGPITR",
    )
    assert PEPTIDES.equals(calc.dataset)
    assert METADATA.equals(calc.metadata)
    assert list(METADATA.columns) == calc.metadata_list
    assert METADATA.columns[0] == calc.key_metadata


@pytest.mark.parametrize(
    "dataset",
    [
        PEPTIDES,
        "./data/peptides.txt",
        "./data/peptides.csv",
        DATA_PATH / "peptides.txt",
        DATA_PATH / "peptides.csv",
        pytest.param(
            lambda: open(DATA_PATH / "peptides.txt", "r"), id="dataset_txt_text_file"
        ),
        pytest.param(
            lambda: open(DATA_PATH / "peptides.txt", "rb"), id="dataset_txt_binary_file"
        ),
        pytest.param(
            lambda: open(DATA_PATH / "peptides.csv", "r"), id="dataset_csv_text_file"
        ),
        pytest.param(
            lambda: open(DATA_PATH / "peptides.csv", "rb"), id="dataset_csv_binary_file"
        ),
    ],
)
def test_setup_dataset_input_types(dataset):
    if callable(dataset):
        with dataset() as f:
            calc = Calculator(dataset=f)
    else:
        calc = Calculator(dataset=dataset)

    assert calc.dataset is not None
    assert not calc.dataset.empty
    assert "Sequence" in calc.dataset.columns


@pytest.mark.parametrize(
    "metadata",
    [
        METADATA,
        "./data/metadata.csv",
        DATA_PATH / "metadata.csv",
        pytest.param(
            lambda: open(DATA_PATH / "metadata.csv", "r"), id="metadata_text_file"
        ),
        pytest.param(
            lambda: open(DATA_PATH / "metadata.csv", "rb"), id="metadata_binary_file"
        ),
    ],
)
def test_setup_metadata_input_types(metadata):
    if callable(metadata):
        with metadata() as f:
            calc = Calculator(metadata=f)
    else:
        calc = Calculator(metadata=metadata)

    assert calc.metadata is not None
    assert not calc.metadata.empty


def test_set_feature_params():
    calc = Calculator()
    calc.set_feature_params(gravy=True)
    assert ("gravy", True) in calc.feature_params.items()
    assert ("molecular_weight", False) in calc.feature_params.items()
    assert "self" not in calc.feature_params.keys()


def test_set_plot_params():
    calc = Calculator()
    calc.set_plot_params(hydropathy_profile=True)
    assert ("hydropathy_profile", True) in calc.plot_params.items()
    assert ("titration_curve", False) in calc.plot_params.items()
    assert "self" not in calc.plot_params.keys()


def test_ensure_attrs():
    calc = Calculator()
    with pytest.raises(ValueError) as e:
        calc.get_features()
    assert "not available" in str(e.value)
    calc.setup(dataset=PEPTIDES, seq="SVIDQSRVLNLGPITR")
    with pytest.raises(ValueError) as e:
        calc.get_plots()
    assert "not available" in str(e.value)


def test_get_features_with_params():
    calc = Calculator(dataset=PEPTIDES, feature_params={"gravy": True})
    res = calc.get_features()
    assert len(PEPTIDES.columns) + 1 == len(res.columns)
    expected = calc.gravy(calc.dataset["Sequence"][0])
    assert expected == res["GRAVY"][0]


def test_get_features_without_params():
    calc = Calculator(dataset=PEPTIDES)
    res = calc.get_features()
    assert len(FEATURES) + len(PEPTIDES.columns) == len(res.columns)


def test_get_peptide_features_with_params():
    calc = Calculator(seq="SVIDQSRVLNLGPITR", feature_params={"gravy": True})
    res = calc.get_peptide_features()
    assert 2 == len(res.columns)
    assert 0.075 == res["GRAVY"][0]


def test_get_peptide_features_without_params():
    calc = Calculator(seq="SVIDQSRVLNLGPITR")
    res = calc.get_peptide_features()
    assert len(FEATURES) + 1 == len(res.columns)


def test_get_plots_for_seq_with_params():
    calc = Calculator(
        seq="SVIDQSRVLNLGPITR",
        plot_params={"hydropathy_profile": True},
    )
    plots = calc.get_plots()
    assert 1 == len(plots)
    assert isinstance(plots[0], go.Figure)


def test_get_plots_for_dataset_with_params():
    calc = Calculator(
        dataset=PEPTIDES,
        metadata=METADATA,
        feature_params={"molecular_weight": True},
        plot_params={
            "raincloud": True,
            "raincloud_feature": "Molecular weight",
            "raincloud_group_by": "Group",
            "raincloud_log_scaled": True,
        },
    )
    calc.get_features()
    plots = calc.get_plots()
    assert 1 == len(plots)
    assert isinstance(plots[0], go.Figure)


def test_get_plots_without_params():
    calc = Calculator(dataset=PEPTIDES, metadata=METADATA, seq="SVIDQSRVLNLGPITR")
    calc.get_features()
    plots = calc.get_plots()
    assert len(PLOTS) == len(plots)


def test_get_plots_as_tuple():
    calc = Calculator(dataset=PEPTIDES, metadata=METADATA, seq="SVIDQSRVLNLGPITR")
    calc.get_features()
    plots = calc.get_plots(as_tuple=True)
    assert 2 == len(plots)



import pytest

from pepmeasure.features import (
    _seq_length,
    _aa_frequency,
    _gravy,
    _isoelectric_point,
    _molecular_formula,
    _molecular_weight,
    _mass,
    _theoretical_mz,
    _one_letter_code,
    _three_letter_code,
    _compute_features,
    _aromaticity,
    _aa_classification,
    _charge_at_ph,
    _charge_density,
    _boman_index,
    _aliphatic_index,
    _extinction_coefficient,
    _instability_index,
)
from tests.constants import PEPTIDES

# Any function that calls one of these functions is already covered by a test for invalid amino acids.
INVALID_SEQ = "ABC"


@pytest.mark.parametrize(
    "func, seq",
    [
        (_seq_length, INVALID_SEQ),
        (_aa_frequency, INVALID_SEQ),
        (_three_letter_code, INVALID_SEQ),
    ],
)
def test_invalid_amino_acid(func, seq):
    with pytest.raises(ValueError) as e:
        func(seq)
    assert "Invalid amino acid symbol" in str(e.value)


def test_seq_length():
    # Benchmark values from ExPASy (Gasteiger et al., 2005)
    assert 20 == _seq_length("ACDEFGHIKLMNPQRSTVWY")
    assert 50 == _seq_length("LHVEDNDEGSPMYMTRCVAWEHITINTNKHYQLYIMWRDGMWYDRMIPAQ")


def test_aa_frequency():
    # Benchmark values from ExPASy (Gasteiger et al., 2005)
    freq = _aa_frequency("AAACCDEEFFF")
    assert {
        "A": 3,
        "C": 2,
        "D": 1,
        "E": 2,
        "F": 3,
        "G": 0,
        "H": 0,
        "I": 0,
        "K": 0,
        "L": 0,
        "M": 0,
        "N": 0,
        "P": 0,
        "Q": 0,
        "R": 0,
        "S": 0,
        "T": 0,
        "V": 0,
        "W": 0,
        "Y": 0,
    } == freq
    assert 20 == len(freq)


def test_molecular_weight():
    # Benchmark values from ExPASy (Gasteiger et al., 2005)
    assert pytest.approx(799.83) == _molecular_weight("PEPTIDE")
    assert pytest.approx(5724.67, rel=1e-3) == _molecular_weight(
        "AGSCCDCILIQNNADMDTDYVCGLVTQMRHGVLEPHILWWAIMWSCHEMI"
    )


def test_mass():
    assert type(_mass("PEPTIDE", "monoisotopic")) is float
    assert type(_mass("PEPTIDE", "average")) is float
    with pytest.raises(ValueError) as e:
        _mass("PEPTIDE", "foo")
    assert "Unknown option" in str(e.value)


def test_theoretical_mz():
    seq = "PNYSGIRERCAMPWNWSQTTNGHLSEIDPFCEMNVVCCIGMDYHYCCKTW"
    charges = [2, 3]
    res = _theoretical_mz(seq, charges)
    res_list = [float(val) for val in res.split(", ")]
    assert isinstance(res_list, list)
    assert len([2, 3]) == len(res_list)


def test_three_letter_code():
    # Benchmark values from Sequence Manipulation Suite (Stothard, 2000)
    assert "ProGluProThrIleAspGlu" == _three_letter_code("PEPTIDE")
    assert (
        "LeuTrpTrpTyrPheMetLysProGluLysLeuAlaGlyGluAsnLysGluProLeuGlnMetMetIleHisTyrIleTyrHisValCysCysTrpAsnGluPheGlyCysAspProGlyValGluLysPheArgProGluMetAlaLeu"
        == _three_letter_code("LWWYFMKPEKLAGENKEPLQMMIHYIYHVCCWNEFGCDPGVEKFRPEMAL")
    )


def test_one_letter_code():
    # Benchmark values from Sequence Manipulation Suite (Stothard, 2000)
    assert "PEPTIDE" == _one_letter_code("ProGluProThrIleAspGlu")
    assert "YLCSIKSTPPLVFGQVDNVHFCMEIPKSFDVRENSRWVDDALEFVYYQVG" == _one_letter_code(
        "TyrLeuCysSerIleLysSerThrProProLeuValPheGlyGlnValAspAsnValHisPheCysMetGluIleProLysSerPheAspValArgGluAsnSerArgTrpValAspAspAlaLeuGluPheValTyrTyrGlnValGly"
    )
    with pytest.raises(ValueError) as e:
        _one_letter_code("Pro Glu Pro Thr Ile Asp Glu")
    assert "Invalid input" in str(e.value)
    with pytest.raises(ValueError) as e:
        _one_letter_code("PrGluProThrIleAspGlu")
    assert "Invalid three letter code" in str(e.value)


def test_gravy():
    # Benchmark values from ExPASy (Gasteiger et al., 2005)
    assert pytest.approx(-1.414) == _gravy("PEPTIDE")
    assert pytest.approx(-0.744) == _gravy(
        "ENFNDTHIIVINCNHVCAECRDTPGWHKCKVPIRMQQMRKWPAESNTRYI"
    )


def test_molecular_formula():
    # Benchmark value from Nomenclature and symbolism for amino acids and peptides (IUPAC et al., 1984)
    assert "C5H9NO4" == _molecular_formula("E")
    # Benchmark values from ExPASy (Gasteiger et al., 2005)
    assert "C34H53N7O15" == _molecular_formula("PEPTIDE")
    assert "C266H401N69O78S5" == _molecular_formula(
        "WQNTDTSMIESSPIGHKDHRTLPTYQWERCWGKSVMELIVCSIWTLYICE"
    )


def test_isoelectric_point():
    assert type(_isoelectric_point("PEPTIDE", "kozlowski")) is float
    assert type(_isoelectric_point("PEPTIDE", "bjellqvist")) is float
    with pytest.raises(ValueError) as e:
        _isoelectric_point("PEPTIDE", "foo")
    assert "Unknown option" in str(e.value)


def test_compute_features_with_dataset():
    params = {"gravy": True}
    res = _compute_features(df=PEPTIDES, params=params)
    assert "GRAVY" in res.columns
    assert "Molecular weight" not in res.columns
    res_grouped = res.groupby("Sequence")["GRAVY"].nunique()
    assert (res_grouped <= 1).all()


def test_compute_features_with_seq():
    params = {"gravy": True}
    res = _compute_features(seq="PEPTIDE", params=params)
    assert "GRAVY" in res.columns
    assert "Molecular weight" not in res.columns
    assert len(res) == 1


def test_aromaticity():
    # Benchmark values from Sequence Manipulation Suite (Stothard, 2000)
    assert pytest.approx(0.0) == _aromaticity("PEPTIDE")
    assert pytest.approx(0.08) == _aromaticity(
        "PKMMDHQPIKTYWCMIGKPNREEIEIAKKMMAEMTDNDWPLHQMPFCSKL"
    )


def test_aa_classification():
    # Benchmark values from (Pommié et al., 2004)
    assert {
        "Aliphatic": 7,
        "Sulfur": 4,
        "Hydroxyl": 9,
        "Basic": 6,
        "Acidic": 2,
        "Amide": 4,
        "Other": 18,
    } == _aa_classification(
        "FIHIPNAWWGADCWCRTWRMQPKSWVFFSQTGAWTFPCPESIKTKTSWNP", "chemical"
    )
    assert {
        "Non-polar": 29,
        "Uncharged": 13,
        "Charged": 8,
    } == _aa_classification(
        "FIHIPNAWWGADCWCRTWRMQPKSWVFFSQTGAWTFPCPESIKTKTSWNP", "charge"
    )
    with pytest.raises(ValueError) as e:
        _aa_classification("PEPTIDE", "foo")
    assert "Unknown option" in str(e.value)


def test_charge_at_ph():
    assert type(_charge_at_ph("PEPTIDE", 7.0)) is float


def test_charge_density():
    # Benchmark values from modlAMP (Müller et al., 2017)
    assert pytest.approx(-0.00375, abs=1e-5) == _charge_density("PEPTIDE", 7.0)
    assert pytest.approx(0.00036, abs=1e-5) == _charge_density(
        "LWSKKWMGGTQDRDVACGHFGKMWILEDTQLGSEKGLSSNTRSYRYQQHP", 7.0
    )


def test_boman_index():
    assert type(_boman_index("PEPTIDE")) is float


def test_aliphatic_index():
    # Benchmark values from ExPASy (Gasteiger et al., 2005)
    assert pytest.approx(55.71) == _aliphatic_index("PEPTIDE")
    assert pytest.approx(70.20) == _aliphatic_index(
        "DPTWFWLEFSLYEERSMDGAPGDGLYFQDDMLDFCLKQKINIVWHRYLKY"
    )


def test_extinction_efficient():
    # Benchmark values from ExPASy (Gasteiger et al., 2005)
    assert 0 == _extinction_coefficient("PEPTIDE", oxidized=False)
    assert 34490 == _extinction_coefficient(
        "HLDQWLPALKQLWLRMAIMWMTWMHDPLPCLNWSMCLQGIWATKNYASGQ", oxidized=False
    )
    assert 34615 == _extinction_coefficient(
        "HLDQWLPALKQLWLRMAIMWMTWMHDPLPCLNWSMCLQGIWATKNYASGQ", oxidized=True
    )


def test_instability_index():
    assert type(_instability_index("PEPTIDE")) is float



import pandas as pd
import plotly.io as pio
from pathlib import Path

from pepmeasure.features import (
    _compute_features,
    _seq_length,
    _aa_frequency,
    _molecular_weight,
    _mass,
    _theoretical_mz,
    _three_letter_code,
    _one_letter_code,
    _gravy,
    _molecular_formula,
    _isoelectric_point,
    _aromaticity,
    _aa_classification,
    _charge_at_ph,
    _charge_density,
    _boman_index,
    _aliphatic_index,
    _extinction_coefficient,
    _instability_index,
)
from pepmeasure.plots import (
    PLOTS,
    _generate_plots,
    _aa_distribution,
    _hydropathy_profile,
    _classification,
    _titration_curve,
    _compare_features,
    _compare_feature,
    _raincloud,
    _mann_whitney_u_test,
)
from pepmeasure.constants import COLORS, PROJECT_PATH, DATA_PATH, ALLOWED_INPUT_TYPES
from pepmeasure.maxquant import _convert_peptide_txt


class Calculator:
    """
    The central interface for using the PepMEASURE library. Computes peptide-specific of dataset-specific features and plots based on defined parameters.
        dataset: pandas DataFrame, path or file object as .txt or .csv. Contains the peptidomic data. The column 'Sequence' must contain the amino acid sequences. To link the metadata file, the first metadata column must be added to the dataset.
        metadata: pandas DataFrame, path or file object as .csv. Contains the metadata. The first column must contain unique identifiers (used as the key). All other columns can provide additional information for each key (e.g., group, batch, ...).
        seq: Amino acid sequence of interest
        feature_params: Dictionary containing all available features and their associated parameters. Use set_feature_params() separately to get an overview on all options.
        plot_params: Dictionary containing all available plots and their associated parameters. Use set_plot_params() separately to get an overview on all options.
    """

    dataset: ALLOWED_INPUT_TYPES
    metadata: ALLOWED_INPUT_TYPES
    metadata_list: list[str]
    key_metadata: str
    seq: str
    feature_params: dict
    plot_params: dict
    computed_features: pd.DataFrame

    def __init__(
        self,
        dataset: ALLOWED_INPUT_TYPES = None,
        metadata: ALLOWED_INPUT_TYPES = None,
        seq: str = None,
        feature_params: dict = None,
        plot_params: dict = None,
    ):
        self.dataset = None
        self.metadata = None
        self.seq = None
        self.setup(
            dataset=dataset,
            metadata=metadata,
            seq=seq,
        )
        self.feature_params = feature_params
        self.plot_params = plot_params
        self.computed_features = None

    # Setup
    def setup(
        self,
        dataset: ALLOWED_INPUT_TYPES = None,
        metadata: ALLOWED_INPUT_TYPES = None,
        seq: str = None,
    ):
        """
        Sets up relevant input data for computing features or generating plots.
            dataset: pandas DataFrame, path or file object as .txt or .csv. Contains the peptidomic data. The column 'Sequence' must contain the amino acid sequences. To link the metadata file, the first metadata column must be added to the dataset.
            metadata: pandas DataFrame, path or file object as .csv. Contains the metadata. The first column must contain unique identifiers (used as the key). All other columns can provide additional information for each key (e.g., group, batch, ...).
            seq: Amino acid sequence of interest
        """
        # TODO Move this logic to a utils function
        if dataset is not None:
            # pandas DataFrame
            if isinstance(dataset, pd.DataFrame):
                self.dataset = dataset
            # Path
            elif isinstance(dataset, (str, Path)):
                path_str = str(dataset).lower()
                if path_str.endswith(".txt"):
                    self.dataset = _convert_peptide_txt(dataset)
                elif path_str.endswith(".csv"):
                    self.dataset = pd.read_csv(dataset)
                else:
                    raise ValueError(
                        f"Unsupported file format. Please provide the dataset as .txt or .csv file, or as a pandas DataFrame."
                    )
            # File object
            else:
                if hasattr(dataset, "name") and dataset.name.lower().endswith(".txt"):
                    self.dataset = _convert_peptide_txt(dataset)
                elif hasattr(dataset, "name") and dataset.name.lower().endswith(".csv"):
                    self.dataset = pd.read_csv(dataset)
                else:
                    raise ValueError(
                        f"Error while setup. Please provide the dataset as .txt or .csv file, or as a pandas DataFrame."
                    )
        if metadata is not None:
            # pandas DataFrame
            if isinstance(metadata, pd.DataFrame):
                self.metadata = metadata
            # Path
            elif isinstance(metadata, (str, Path)):
                path_str = str(metadata).lower()
                if path_str.endswith(".csv"):
                    self.metadata = pd.read_csv(metadata)
                else:
                    raise ValueError(
                        f"Unsupported file format. Please provide the metadata as .csv file or pandas DataFrame."
                    )
            # File object
            elif hasattr(metadata, "name") and metadata.name.lower().endswith(".csv"):
                self.metadata = pd.read_csv(metadata)
            else:
                raise ValueError(
                    f"Error while setup. Please provide the metadata as .csv file or pandas DataFrame."
                )

            self.metadata_list = list(self.metadata.columns)
            self.key_metadata = self.metadata.columns[0]
        if seq is not None:
            self.seq = seq

    def set_feature_params(
        self,
        three_letter_code: bool = False,
        molecular_formula: bool = False,
        seq_length: bool = False,
        molecular_weight: bool = False,
        mass: bool = False,
        mass_option: str = "monoisotopic",
        theoretical_mz: bool = False,
        gravy: bool = False,
        isoelectric_point: bool = False,
        isoelectric_point_option: str = "bjellqvist",
        aromaticity: bool = False,
        charge_at_ph: bool = False,
        charge_at_ph_level: float = 7.0,
        charge_density: bool = False,
        charge_density_level: float = 7.0,
        boman_index: bool = False,
        aliphatic_index: bool = False,
        extinction_coefficient: bool = False,
        extinction_coefficient_oxidized: bool = False,
        instability_index: bool = False,
    ):
        """
        Selects peptide features and their related parameters.
        """
        params = locals().copy()
        params.pop("self")
        self.feature_params = params

    def set_plot_params(
        self,
        aa_distribution: bool = False,
        aa_distribution_order_by: str = "frequency",
        aa_distribution_show_all: bool = False,
        hydropathy_profile: bool = False,
        classification: bool = False,
        classification_classify_by: str = "chemical",
        titration_curve: bool = False,
        compare_features: bool = False,
        compare_features_a: str = "Sequence length",
        compare_features_b: str = "Molecular weight",
        compare_features_group_by: str = None,
        compare_features_intensity_threshold: float = None,
        compare_feature: bool = False,
        compare_feature_a: str = "GRAVY",
        compare_feature_group_by: str = None,
        compare_feature_intensity_threshold: float = None,
        raincloud: bool = False,
        raincloud_feature: str = "GRAVY",
        raincloud_group_by: str = None,
        raincloud_log_scaled: bool = True,
        mann_whitney: bool = False,
        mann_whitney_feature: str = "GRAVY",
        mann_whitney_group_by: str = None,
        mann_whitney_group_a: str = None,
        mann_whitney_group_b: str = None,
        mann_whitney_alternative: str = "two-sided",
        mann_whitney_intensity_threshold: float = None,
    ):
        """
        Selects peptide and dataset plots and their related parameters.
        """
        params = locals().copy()
        params.pop("self")
        self.plot_params = params

    # Utils
    def _ensure_attrs(self, *attrs):
        """
        Raises an error if a given attribute wasn't defined before.
        """
        missing = [a for a in attrs if getattr(self, a) is None]
        if missing:
            if "computed_features" in missing:
                msg = "Computed features are not available. Feature visualisation requires that get_features() has been executed first."
            else:
                msg = f"The following information is not available: {missing}. Please execute the corresponding set or get methods first."
            raise ValueError(msg)

    # Features
    def get_features(self) -> pd.DataFrame:
        """
        Computes selected features on the current dataset. Requires a dataset set by setup().
        Note: If no features were explicitly selected, all available features are computed with their default options.
        """
        self._ensure_attrs("dataset")
        if self.feature_params:
            params = self.feature_params
        else:
            params = {"select_all": True}
        self.computed_features = _compute_features(
            params=params,
            df=self.dataset,
            seq=None,
        )
        return self.computed_features

    def get_peptide_features(self) -> pd.DataFrame:
        """
        Computes selected features on the current peptide sequence of interest. Requires a sequence set by setup().
        Note: If no features were explicitly selected, all available features are computed with their default options.
        """
        self._ensure_attrs("seq")
        if self.feature_params:
            params = self.feature_params
        else:
            params = {"select_all": True}
        return _compute_features(
            params=params,
            df=None,
            seq=self.seq,
        )

    seq_length = staticmethod(_seq_length)
    aa_frequency = staticmethod(_aa_frequency)
    molecular_weight = staticmethod(_molecular_weight)
    mass = staticmethod(_mass)
    theoretical_mz = staticmethod(_theoretical_mz)
    three_letter_code = staticmethod(_three_letter_code)
    one_letter_code = staticmethod(_one_letter_code)
    gravy = staticmethod(_gravy)
    molecular_formula = staticmethod(_molecular_formula)
    isoelectric_point = staticmethod(_isoelectric_point)
    aromaticity = staticmethod(_aromaticity)
    aa_classification = staticmethod(_aa_classification)
    charge_at_ph = staticmethod(_charge_at_ph)
    charge_density = staticmethod(_charge_density)
    boman_index = staticmethod(_boman_index)
    aliphatic_index = staticmethod(_aliphatic_index)
    extinction_coefficient = staticmethod(_extinction_coefficient)
    instability_index = staticmethod(_instability_index)

    # Plots
    def get_plots(
        self,
        as_tuple: bool = False,
        colors: list = COLORS,
    ) -> list | tuple:
        """
        Generates selected plots. Requires a sequence or a dataset set by setup().
        Note: If no plots were explicitly selected, all available plots are computed with their default options.
            as_tuple: If set to True, the peptide and dataset plots are returned seperated as tuple.
        """
        if colors is None:
            colors = COLORS

        if self.plot_params:
            params = self.plot_params
        else:
            params = {"select_all": True}

        selected = [
            feature
            for key, feature in PLOTS.items()
            if params.get(key) or params.get("select_all")
        ]
        has_seq_based = any(f.seq_based for f in selected)
        has_dataset_based = any(not f.seq_based for f in selected)

        if has_seq_based:
            self._ensure_attrs("seq")
        if has_dataset_based:
            self._ensure_attrs("computed_features", "metadata")
            current_features = pd.merge(
                self.computed_features, self.metadata, on=self.key_metadata, how="left"
            )
        else:
            current_features = self.dataset

        plot_tuple = _generate_plots(
            df=current_features,
            seq=self.seq,
            params=params,
            colors=colors,
        )
        if as_tuple:
            return plot_tuple
        else:
            plots = [plot for sublist in plot_tuple for plot in sublist]
            return plots

    aa_distribution = staticmethod(_aa_distribution)
    hydropathy_profile = staticmethod(_hydropathy_profile)
    classification = staticmethod(_classification)
    titration_curve = staticmethod(_titration_curve)
    compare_features = staticmethod(_compare_features)
    compare_feature = staticmethod(_compare_feature)
    raincloud = staticmethod(_raincloud)
    mann_whitney_u_test = staticmethod(_mann_whitney_u_test)

    # Demonstration: Hello PepMEASURE!
    @staticmethod
    def hello_pepmeasure():
        print("✨ Hello PepMEASURE! ✨")
        # Load data
        dataset = pd.read_csv(DATA_PATH / "peptides.csv")
        metadata = pd.read_csv(DATA_PATH / "metadata.csv")
        print("Loaded input data.")

        # Initialize Calculator instance
        calc = Calculator(
            dataset=dataset,
            metadata=metadata,
            seq="SVIDQSRVLNLGPITR",
        )
        print("Set up calculator.")

        # Select features and plots
        calc.set_feature_params(
            gravy=True,
            molecular_weight=True,
        )
        calc.set_plot_params(
            hydropathy_profile=True,
            classification=True,
            classification_classify_by="charge",
        )
        print("Set feature and plot parameters.")

        # Compute and output results
        print("Printing first five peptides with computed features ...")
        print(calc.get_features().head())
        plots = calc.get_plots()
        i = 1
        for plot in plots:
            pio.write_image(plot, PROJECT_PATH / "results" / f"plot{i}.png", scale=3)
            i += 1
        results_path = PROJECT_PATH / "results"
        print(f"You can find all generated plots in '{results_path}'.")
        print("Successfully executed PepMEASURE demonstration. Happy coding! 💻🧬")



# This logic is based on the implementation from
# PROTzilla (https://github.com/cschlaffner/PROTzilla/tree/dev/backend/protzilla/importing)
# and has been adapted to fit this project.

import re
from enum import Enum
import pandas as pd

from .constants import ALLOWED_INPUT_TYPES


def _read_peptide_file(file):
    if hasattr(file, "read"):
        file.seek(0)
    return pd.read_csv(
        file,
        sep="\t",
        low_memory=False,
        na_values=["", 0],
        keep_default_na=True,
    )


def _convert_peptide_txt(
    file: ALLOWED_INPUT_TYPES, intensity_name: str = "Intensity"
) -> pd.DataFrame:
    """
    Converts a MaxQuant peptide.txt file into a pandas DataFrame.
        file: File object (e.g., uploaded file) or path to the peptide.txt file.
        intensity_name: Name of the intensity column to extract. Default is "Intensity".
    """
    allowed = {item.value for item in IntensityType}
    if intensity_name not in allowed:
        raise ValueError(f"Unknown intensity name: {intensity_name}")

    if intensity_name in [IntensityType.LFQ_INTENSITY.value, IntensityType.IBAQ.value]:
        intensity_name = IntensityType.INTENSITY.value

    id_columns = [
        "Leading razor protein",
        "Sequence",
        "Missed cleavages",
        "PEP",
        "Charges",
    ]

    df = _read_peptide_file(file)

    if "Sample" not in df.columns:
        id_df = df[id_columns]
        disallowed_suffixes = r"(variability|count|type|peptides)"
        if intensity_name in (
            IntensityType.RATIO_HL.value,
            IntensityType.RATIO_LH.value,
        ):
            base_pattern = rf"^{re.escape(intensity_name)}\s(?!normalized\b)(?!.*\b{disallowed_suffixes}\b).*$"
        else:
            base_pattern = (
                rf"^{re.escape(intensity_name)}\s(?!.*\b{disallowed_suffixes}\b).*$"
            )

        intensity_df = df.filter(regex=base_pattern, axis=1)
        intensity_df.columns = [
            c[len(intensity_name) + 1 :] for c in intensity_df.columns
        ]
        molten = pd.melt(
            pd.concat([id_df, intensity_df], axis=1),
            id_vars=id_columns,
            var_name="Sample",
            value_name="Intensity",
        )

    else:
        final_df = df.rename(columns={"Proteins": "Protein ID"})
        ordered = final_df[
            [
                "Sample",
                "Protein ID",
                "Sequence",
                "Intensity",
                "PEP",
                "Charges",
            ]
        ]

    molten = molten.rename(columns={"Leading razor protein": "Protein ID"})
    ordered = molten[
        [
            "Sample",
            "Protein ID",
            "Sequence",
            "Intensity",
            "PEP",
            "Charges",
        ]
    ].copy()
    ordered.dropna(subset=["Protein ID"], inplace=True)
    ordered.sort_values(by=["Sample", "Protein ID"], ignore_index=True, inplace=True)

    new_groups, filtered_proteins = _clean_protein_groups(
        ordered["Protein ID"].tolist()
    )
    cleaned = ordered.assign(**{"Protein ID": new_groups})

    return cleaned


def _clean_protein_groups(protein_groups: list):
    regex = {
        "ensembl_peptide_id": re.compile(r"^ENSP\d{11}"),
        "refseq_peptide": re.compile(r"^NP_\d{6,}"),
        "refseq_peptide_predicted": re.compile(r"^XP_\d{9}"),
    }
    uniprot_regex = re.compile(
        r"""^[A-Z]               # start with capital letter
        [A-Z\d]{5}([A-Z\d]{4})?  # match ids of length 6 or 10
        ([-_][-\d]+)?            # match variations like -8 and _9-6
        """,
        re.VERBOSE,
    )

    removed_protein_ids = []

    extracted_ids = {k: set() for k in regex.keys()}
    found_ids_per_group = []
    # go through all groups and find the valid proteins
    # non uniprot ids are put into extracted_ids, so they can be mapped
    for group in protein_groups:
        found_in_group = []
        for protein_id in group.split(";"):
            if not protein_id.startswith("ENSP") and (
                match := uniprot_regex.search(protein_id)
            ):
                found_in_group.append(match.group(0))
                continue
            for identifier, pattern in regex.items():
                if match := pattern.search(protein_id):
                    found_id = match.group(0)
                    extracted_ids[identifier].add(found_id)
                    found_in_group.append(found_id)
                    break  # can only match one regex
            else:
                removed_protein_ids.append(protein_id)
        found_ids_per_group.append(found_in_group)

    new_groups = []
    for group in found_ids_per_group:
        all_ids_of_group = []
        for old_id in group:
            if uniprot_regex.search(old_id):
                all_ids_of_group.append(old_id)
            # elif map_to_uniprot:
            # [ Removed ]
            else:
                all_ids_of_group.append(old_id)
        new_groups.append(all_ids_of_group[0] if all_ids_of_group else "")
    return new_groups, removed_protein_ids


class IntensityType(Enum):
    IBAQ = "iBAQ"
    INTENSITY = "Intensity"
    LFQ_INTENSITY = "LFQ intensity"
    RATIO_HL = "Ratio H/L"
    RATIO_LH = "Ratio L/H"
    RATIO_HL_NORMALIZED = "Ratio H/L normalized"
    RATIO_LH_NORMALIZED = "Ratio L/H normalized"
